require "confluence_helper/version"

module ConfluenceHelper
  
  class HtmlTable
    def initialize(node)
      @node = node
    end
    
    def has_columns?(columns:)
      column_names.sort == columns.sort
    end
    
    def column_names
      @column_names ||= @node.xpath('.//tr/th').map{|n| n.text.strip}
    end
    
    def data
      @data ||= 
        @node
          .xpath('.//tr')
          .map{|r| r.xpath('.//td').map{|cell| cell_value(cell)} }
          .reject{|r| r.empty? }
          .map{|r| column_names
                      .each
                        .with_index
                          .with_object({}){ |(c,i),o| o[c] = r[i]  }
              }
    end
    
    private
    
    def cell_value(cell)
      (cell.xpath(".//a").empty? ? cell.text : cell.inner_html).gsub("\u00C2", " ").gsub("\u00A0", " ").strip
    end
  end


class Confluence

  def get_labels_for_each_child_page(rootpageid)
    child_pages_in_space(rootpageid, {:expand => "metadata.labels"} ).
      each_with_object({}) do |page, hash|
        hash[page["title"]] = 
          page.
            fetch("metadata", {}).
            fetch("labels",{}).
            fetch("results",[]).
            map{|l| l["name"]}
      end
  end
  
  def find_url_by_title_and_space(spacekey, title)
    page_by_title_in_space(spacekey, title, {})["_links"]["self"]
  end 
  
  def save_page_attachment( spacekey, page_title, attachment_name, destination )
    response = query_rest_api( get_attachment_url( spacekey, page_title, attachment_name ) )
    File.open(destination, 'wb') { |fp| fp.write(response.body) }
  end
  
  def deliverables_pages(rootpageid)
    child_pages_in_space(rootpageid, {:expand => "metadata.labels,body.export_view"} ).
      each_with_object({}) do |page, hash|
        hash[page["title"]] = DeliverablePage.new( 
            labels: page.fetch("metadata", {}).fetch("labels", {}).fetch("results",[]).map{|l| l["name"]},
            body: page.fetch("body", {}).fetch("export_view", {}).fetch("value", "")
         )
      end
  end
  
  private

  @@url = {}
  
  class DeliverablePage
    def initialize(labels:, body:)
      @node = Nokogiri::HTML(body)
    end
    
    def tables
      @tables ||= @node.xpath('//table').map{|t| HtmlTable.new(t) }
    end

    def deliverables
      tables.select{|t| is_deliverables_table? t } .map{|table| table.data }.flatten
    end

    #Name DHFR  Status  Location  Comments
    def is_deliverables_table?(table)
      table.has_columns? columns: %w{Name DHFR Status Location Comments}
    end

  end
 
  def get_attachment_url(spacekey, page_title, attachment_name)
    "#{ENVied.CONSTASH_URL}#{ get_attachment_data(spacekey, page_title, attachment_name)["_links"]["download"]}"
  end
  
  def get_attachment_data(spacekey, page_title, attachment_name)
    get_attachments_data(spacekey, page_title).find{|x| x["title"] == attachment_name}
  end
 
  def get_attachments_data(spacekey, page_title)
    page_by_title_in_space(spacekey, page_title, {expand:"children.attachment"})["children"]["attachment"]["results"]
  end
  
  def page_by_title_in_space(spacekey, page_title, options)
    all_pages = pages_in_space(spacekey, options).lazy
    all_pages.find {|p| p['title'] == page_title}.
      tap{|page| raise "Can't find page with title #{page_title} in that space #{spacekey}\n\tFound pages: #{all_pages.map{|p| p['title']}.join("\n\t")}" unless page }
  end   
  
  def pages_in_space(spacekey, options)
    enum_for(:for_all_pages_of_results, "rest/api/space/#{spacekey}/content/page", options)
  end  
  
  def child_pages_in_space(pageid, options)
    enum_for(:for_all_pages_of_results, "rest/api/content/#{pageid}/child/page", options)
  end  
  
  def for_all_pages_of_results(url, options, &block)
    full_options = {start:0, limit:25}.merge(options)
    response = query_rest_api(url, full_options)
    if response && response.success? 
      results = response.body.fetch('results',[])
      results.each{ |p| block.call(p) }
      if(results.count == full_options[:limit])
        start = full_options[:start] + results.count 
        for_all_pages_of_results(url, options.merge({start:start}), &block)  
      end
    end
  end  
  
  def get_logged_in_connection(request = :json)
    # TODO: move to memcache
    # store = ActiveSupport::Cache.lookup_store(:mem_cache_store, ['localhost:11211'])
  
    return Faraday.new(:url => ENVied.CONSTASH_URL) do |faraday|
      faraday.use :http_cache # , store: store
      faraday.request :url_encoded              # form-encode POST params
      faraday.basic_auth(ENVied.CONSTASH_USERNAME, ENVied.CONSTASH_PASSWORD)
      faraday.request request

      faraday.response :json, :content_type => /\bjson$/
#      faraday.response :logger                  # log requests to STDOUT

      faraday.adapter  Faraday.default_adapter  # make requests with Net::HTTP
    end
  end
  
  def query_rest_api(url, options = {})
    puts "-->#{url} #{options.inspect}"
    get_logged_in_connection().
      get(url, options).
        tap{ |response| 
          raise "Error while attempting to query '#{ENVied.CONSTASH_URL}/#{url}'\n Status: #{response.status} \n #{response.body}" unless [200, 302].include? response.status  
        }
  end
  
  def get_url(space, title)
    @@url[ [space,title] ] ||= find_url_by_title_and_space(space, title)
  end
end

class WikiPage < Confluence

  def initialize(spacekey, title)
    @spacekey, @title = spacekey, title
  end
  
  def to_nodes(mode=:text)
    parse_wiki_table(doc, mode)
  end
  
  def attachment(title)
    page = query_rest_api( url , {expand:"children.attachment"})
    results = page.body["children"]["attachment"]["results"]
    result = results.find{|x| x["title"] == title}
    attachment_url = "#{ENVied.CONSTASH_URL}#{result["_links"]["download"]}"
    query_rest_api(attachment_url, options = {}).body
  end
  
  private 
  
  def url
    @url ||= get_url(@spacekey, @title)
  end
  
  def doc
    @doc ||= get_page_body()
  end
  
  def get_page_body()
    page = query_rest_api( url , {expand:"body.export_view"} )
    return Nokogiri::HTML( page.body["body"]["export_view"]["value"] )
  end 
  
  def parse_wiki_table(doc, mode=:text)
    rows = doc.xpath("//table/tbody/tr")
    headers =  rows.xpath("th").map{|n| n.content}
    rows = rows - rows.xpath("th/..")
    documents = rows.map{ | row | row_to_hash(row, headers, mode) }  #convert to hashes
    if headers.include?("ID")
      documents.each_with_object({}){|doc, o| o[doc["ID"].to_i] = doc } # index by id
    else
      documents
    end
  end  
  
  def row_to_hash(row, headers, mode = :text)
    row.xpath( "td" ).
      to_a.
        each.
          with_index.
          with_object( {} ) { | (cell , index) , row_hash | 
            row_hash[ headers[ index ] ] = 
              (mode == :text ? cell.content : cell.inner_html).gsub( /\u00A0/,"" )
          }
  end
  
end

end
